
/*
Copyright [2015] [O.Klemke]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/* 
pi_power shut down the system when the gpio-pin 0 ( change it in the header-file )
goes from HIGHT to LOW
*/


#include "pi_powerbtn.h"

// globals
extern int errno ;
static volatile int err = 0;
static volatile int PwrButton__ = PwrButton ;
static volatile int LiveLed__ = LiveLed ;
static volatile int PwrButton_delay__ = PwrButton_delay;
static volatile int Daemon__ = Daemon;

static volatile int pwr_pidfile = -1 ; // this process had have created the pid file or
static volatile int pwr_lockfile = -1 ; // lock file important for cleaning up
static volatile int led_lockfile = -1 ; // lock file important for cleaning up
char pidfile[128] ;
char lockfile[128];
char LED_lockfile[128];



// We leave the system clean
void clean_end(int err_end)
{

if ( pwr_lockfile >= 0 )
    {
    pwr_lockfile = -1 ;  // signal handlers are asyncron
    unlink( lockfile ); // cleaning up lock file
    }; 

if ( led_lockfile >= 0 )
    {
    led_lockfile = -1 ;  // signal handlers are asyncron
    unlink( LED_lockfile ); // cleaning up lock file
    }; 


if ( pwr_pidfile >= 0 )
    {
    pwr_pidfile = -1 ;  // signal handlers are asyncron
    remove( pidfile ); // cleaning up pid file
    };





exit(err_end);
};




// we will read the commandline, if there are arguments

int PwrBuCml(int argc, char *argv[])
{
    if ( argc >= 20 )
    {
    fprintf(stderr,"pi_powerbtn: too many arguments (%i)\n",argc);
    clean_end(1);
    };

int count = 1 ;
int count_buf ;
int pin_buf;
int tout_buf;
const char version[] = VERSION;

while (argc > count) 
    {
		if ( ( strcasecmp( argv[count], "-p" ) == 0 ) || ( strcasecmp(argv[count], "--pin") == 0 ))
		    {
		    count_buf = count;
		    ++count;
		        if ( strlen(argv[count]) > 3 )
		        {
		        fprintf(stderr,"pi_powerbtn: bad argument (%s %s)\n use %s -h for help \n", argv[count_buf], argv[count], argv[0] );
		        syslog(LOG_ERR,"pi_powerbtn: bad argument (%s %s)\n use %s -h for help \n", argv[count_buf], argv[count], argv[0] );
		        clean_end(-1);
			};
		     pin_buf = atoi(argv[count]);
			if (( pin_buf < 0 ) || ( pin_buf > 29 ))
			{
			    fprintf(stderr,"pi_powerbtn:  unkown wiringPi Pin --pin %i /n check it or look out for an update ",pin_buf);
			    syslog(LOG_ERR,"pi_powerbtn:  unkown wiringPi Pin --pin %i /n check it or look out for an update ",pin_buf);
			    clean_end(-1);
			}; 
		    PwrButton__ = pin_buf; 
		    };
		if ( ( strcasecmp( argv[count], "-l" ) == 0 ) || ( strcasecmp(argv[count], "--led") == 0 ))
		    {
		    count_buf = count;
		    ++count;
		        if ( strlen(argv[count]) > 3 )
		        {
		        fprintf(stderr,"pi_powerbtn: bad argument (%s %s)\n use %s -h for help \n", argv[count_buf], argv[count], argv[0] );
		        syslog(LOG_ERR,"pi_powerbtn: bad argument (%s %s)\n use %s -h for help \n", argv[count_buf], argv[count], argv[0] );
		        clean_end(-1);
			};
		     pin_buf = atoi(argv[count]);
			if (( pin_buf < -1 ) || ( pin_buf > 29 ))
			{
			    fprintf(stderr,"pi_powerbtn:  unkown wiringPi Pin --led %i /n check it or look out for an update ",pin_buf);
			    syslog(LOG_ERR,"pi_powerbtn:  unkown wiringPi Pin --led %i /n check it or look out for an update ",pin_buf);
			    clean_end(-1);
			}; 
		    LiveLed__ = pin_buf; 
		    };
		if ( ( strcasecmp( argv[count], "-t" ) == 0 ) || ( strcasecmp(argv[count], "--timeout") == 0 ))
		    {
		    count_buf = count;
		    ++count;
		        if ( strlen(argv[count]) > 4 )
		        {
		        fprintf(stderr,"pi_powerbtn: bad argument (%s %s)\n use %s -h for help \n", argv[count_buf], argv[count], argv[0] );
		        syslog(LOG_ERR,"pi_powerbtn: bad argument (%s %s)\n use %s -h for help \n", argv[count_buf], argv[count], argv[0] );
		        clean_end(-1);
			};
		     tout_buf = atoi(argv[count]);
			if (( tout_buf < 50 ) || ( tout_buf > 1000 ))
			{
			    fprintf(stderr,"pi_powerbtn:  the parameter %i msec is senseless,\n between 50 and 1000 msec will be accepted (%i msec is default) ",tout_buf,PwrButton_delay__);
			    syslog(LOG_ERR,"pi_powerbtn:  the parameter %i msec is senseless,\n between 50 and 1000 msec will be accepted (%i msec is default) ",tout_buf,PwrButton_delay__);
			    clean_end(-1);
			}; 
		    PwrButton__ = tout_buf; 
		    };
		if ( ( strcasecmp( argv[count], "-D" ) == 0 ) || ( strcasecmp(argv[count], "--daemon") == 0 ))
		    {
		    count_buf = count;
		    ++count;
		        if ( strlen(argv[count]) > 4 )
		        {
		        fprintf(stderr,"pi_powerbtn: bad argument (%s %s)\n use %s -h for help \n", argv[count_buf], argv[count], argv[0] );
		        syslog(LOG_ERR,"pi_powerbtn: bad argument (%s %s)\n use %s -h for help \n", argv[count_buf], argv[count], argv[0] );
		        clean_end(-1);
			};
		     Daemon__ = atoi(argv[count]);
		 
		    };

		if ( ( strcasecmp( argv[count], "-h" ) == 0 ) || ( strcasecmp(argv[count], "--help") == 0 ))
		    {
		        fprintf(stderr,"Copyright [2015] [O.Klemke]\nLicensed under the Apache License, Version 2.0 \n see http://www.apache.org/licenses/LICENSE-2.0\npi_powerbtn shuts down you raspberry Pi when you pull down a gpio pin\n  use: pi_powerbtn --pin 0 --timeout 250 --daemon 1 \n -p\n --pin  range 0 to 29 (default %i) the pin to use for power button  in wiringPi counting \n -l\n --led  range 0 to 29 (default %i) the pin to use for alive led in wiringPi counting ( -1 for off)\n -t\n --timeout   range 50 to 1000msec (default %i msec) the time between two pin reads\n -D\n --daemon     should it run as backround system process >=1 means yes (default %i)\n -v\n --version   for infos\n -h\n --help   for help \n",PwrButton ,LiveLed , PwrButton_delay, Daemon  );
		        clean_end(0);
		    };
		if ( ( strcasecmp( argv[count], "-v" ) == 0 ) || ( strcasecmp(argv[count], "--version") == 0 ))
		    {
		        fprintf(stderr," pi_powerbtn version %s \n\n   Copyright [2015] [O.Klemke]\n Licensed under the Apache License, Version 2.0 \n you may not use this file except in compliance with the License.\n You may obtain a copy of the License at\n\n    http://www.apache.org/licenses/LICENSE-2.0\n\n Unless required by applicable law or agreed to in writing, software\n distributed under the License is distributed on an \"AS IS\" BASIS,\n WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n See the License for the specific language governing permissions and\n limitations under the License.\n",version  );
		        clean_end(0);
		    };



	++count;

  };

return err ;
};

// We will set our own pid-file to prevent a double start.
// There is no locking machanism for wring-Pi yet, so we will implent our own. Look for updates
int reg()
{
int reg_err = 0;
int pidfile_hdl; // in a time before a time, this integer was be called "handle", do you know what an Atari ST was??
FILE *pidfile_des;
const char prog_name[15] = ProgramName ; // our name defined in header file let the precompiler do it
char pin_str[10];

pid_t own_pid;     // our process-id
struct stat lock_stat;

// lock file handling

// let's craft a lock-file name

if ( PwrButton__ < 10 )
{
sprintf( pin_str,"0%i",PwrButton__); // there is no itoa in gcc
}
else
{
sprintf(  pin_str,"%i",PwrButton__); // there is no itoa in gcc
};

strcpy( lockfile, "/run/lock/wiringPi." );
strcat( lockfile, pin_str ); // see above ;-)
strcat( lockfile, ".lck" );

// for led
if ( LiveLed__ >=0 )
    {
    if ( LiveLed__ < 10 )
    {
    sprintf( pin_str,"0%i",LiveLed__); // there is no itoa in gcc
    }
    else
    {
    sprintf(  pin_str,"%i",LiveLed__); // there is no itoa in gcc
    };

    strcpy( LED_lockfile, "/run/lock/wiringPi." );
    strcat( LED_lockfile, pin_str ); // see above ;-)
    strcat( LED_lockfile, ".lck" );
    }

// is there the /run/lock directory ?? if not we will make it
if ( stat("/run/lock", &lock_stat) < 0 )
    {
    mkdir("/run/lock", 1777);
    };

// is there a lockfile for this Pin --> exit
if ( stat( lockfile, &lock_stat) > -1 )
    {
    syslog(LOG_ERR ,"the wiringPi Pin %i is allready in use ",PwrButton__ );
    clean_end(-1);
    };

if ( (LiveLed__ >= 0) & ( stat( LED_lockfile, &lock_stat) > -1 ))
    {
    syslog(LOG_ERR ,"the wiringPi LED-Pin %i is allready in use - ignoring led",PwrButton__ );
    LiveLed__ = -1 ; // you got it
    };





//  ^^^^^^^^^^^



// now we make our pid file
own_pid = getpid() ; 

strcpy( pidfile, "/run/");
strcat( pidfile, prog_name );
strcat( pidfile, ".pid");

pidfile_hdl=open( pidfile, O_CREAT | O_EXCL | O_RDWR , 0644 ); // only to get an error if exists

if ( pidfile_hdl < 0 )
    {
    syslog(LOG_ERR ,"Not able to open pid file %i", pidfile_hdl);
    clean_end(pidfile_hdl);
    };

// all ist fine, lets create the files

    pidfile_des = fdopen(pidfile_hdl, "w");
    fprintf(pidfile_des, "%i\n",(int) own_pid);
    pwr_pidfile = (int) pidfile_des ; // this worth will be forgotten and useless soon but it's >0 and nice for debugging
    fclose(pidfile_des); // we left it all cleaned and closed
    close(pidfile_hdl);

    pwr_lockfile=symlink( pidfile, lockfile); // the lock file is a symbolic link to the pid file
    if (LiveLed__ >= 0)
    {
    led_lockfile=symlink( pidfile, LED_lockfile);
    };


return reg_err;
};


void Shutdown()
{

const char *shtdwn = "/sbin/shutdown";
char *shtdwn_argv[] = { "/sbin/shutdown", "-h", "-P", "now", NULL };
char *shtdwn_environ[] = { NULL };
err = execve(shtdwn, shtdwn_argv, shtdwn_environ);

syslog(LOG_ALERT ,"pi_powerbtn: shutting down the system  "); // it's important, I mean


};



int main(int argc, char *argv[])
{


int blnk = 1 ;


openlog("pi_powerbtn", LOG_NDELAY | LOG_PID | LOG_CONS , LOG_DAEMON ); // we have no terminal, so we talk over syslog


// we will read the commandline, if there are arguments - we are no daemon yet
	if (argc > 1) {
	    err = PwrBuCml(argc , argv );
	    };

//  < wiringPi - init >
err = wiringPiSetup() ;
    if ( err != 0)
    {
    fprintf(stderr,"pi_power: No wiringPiSetup possible"); //  we are no daemon yet
    syslog(LOG_ERR,"pi_power: No wiringPiSetup possible"); // but a message more or less....
    clean_end(err) ;
    };

    if (PwrButton__ ==  LiveLed__ )
    {     
    fprintf(stderr,"pi_power: power button pin %i is as the same as the LED pin %i - ignoring led",PwrButton__,LiveLed__ ); //  we are no daemon yet
    syslog(LOG_WARNING,"pi_power: power button pin %i is as the same as the LED pin %i - ignoring led",PwrButton__,LiveLed__ ); // but a message more or less....
    LiveLed__ = -1;
    };

    if( Daemon__ >=1 )
    {
    daemon(0, 2);  // we go in backround ATENTION this is a kind of fork(), the pid is changing 
    };
    
// the signal handling first, what will we do when we are terminated / killed 
signal( SIGTERM, clean_end);  // we are terminated ( #>kill -15 ....)
signal( SIGKILL, clean_end);  // we are killed     ( #>kill -9  ....)
signal( SIGINT, clean_end);   // we are terminated ( #> [ctrl]+[c]  )
signal( SIGQUIT, clean_end);  // we are terminated ( #> [ctrl]+[q]  )
signal( SIGTSTP, clean_end);  // we are asked to stop, this is senseless here, so we die
signal( SIGHUP, SIG_IGN);    // we are huped, we will ignore it, daily trafic like ;-)
signal( SIGUSR1, SIG_IGN);    // we are light huped, we will ignore it, daily trafic like ;-)




reg();

// some infos
if ( (PwrButton__ == 1) || (LiveLed__ == 1) ) { syslog(LOG_WARNING, " The use of Pin 1 interfers with analog output ");};
if ( (PwrButton__ >= 10) && ( PwrButton__ <= 14) ) { syslog(LOG_INFO, " Pin %i has no pull-up(down) resistor on board ",PwrButton__);};
if ( (PwrButton__ == 15) || ( PwrButton__ == 16 ) || (LiveLed__ == 15) || (LiveLed__ == 16)) { syslog(LOG_WARNING, " Pin %i - remember, you have to disable the seriell console in /boot/commandline.txt ",PwrButton__);};

pinMode(PwrButton__ , INPUT);
if ( LiveLed__ >= 0) { pinMode( LiveLed__, OUTPUT ); };

if ( PwrButton__ < 8 )  // grater Pins have not build-in resistors to aktivate, pins 8 and 9 are pulled up
{
pullUpDnControl(PwrButton__ , PUD_UP);
};

//  </wiringPi - init >

delay ( (uint) PwrButton_delay ); // a little time to do it
    syslog(LOG_INFO, " successful started, waiting for event on pin %i to shutdown the system ",PwrButton__);
    if ( LiveLed__ >= 0) { syslog(LOG_INFO, " blinking on pin %i  ",LiveLed__); } else { syslog(LOG_INFO, " blinking is off (%i)  ",LiveLed__); };

    while ( digitalRead ( PwrButton__ ) == 1 )
    {
    
    if ( LiveLed__ >= 0) { digitalWrite (LiveLed__, blnk) ; blnk = blnk ^ 1; }; // we switch the blink state via exclusive OR ( 1^1=0 , 0^1=1 )
    delay ( (uint) PwrButton_delay__ );
    };
if ( LiveLed__ >= 0) { digitalWrite (LiveLed__, 1); }; // switch it on, it will go off when the Pi is halted  
Shutdown();

clean_end(0);
return err ;
};
